# rankade.api.Endpoint.py
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, MutableMapping, Optional, Union

from enum_tools.documentation import document_enum

if TYPE_CHECKING:
    from .Api import HEADERS, JSON, PARAMS


class Endpoint_Method(Enum):
    """HTTP Request methods [RFC2616 Section 5.1.1](https://www.rfc-editor.org/rfc/rfc2616#section-5.1.1)"""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"
    CONNECT = "CONNECT"


@dataclass
class Endpoint_Mixin:
    """Stores the basic information needed for each endpoint."""

    auth: bool
    """Does the request need authorisation before proceeding."""
    paginated: bool
    """Will the response be paginated."""
    method: Endpoint_Method
    """HTTP Method used in the request."""
    base_path: str
    """URL path to use in request."""


@document_enum
class Endpoint(Endpoint_Mixin, Enum):
    """
    Enum with information needed for each request.
    auth, paginated, method, base_path
    """

    AUTH = False, False, Endpoint_Method.GET, "auth"
    """
    Authorisation returns JWT. [GET /auth](https://rankade.com/api/#get-auth)
    """
    GAME = True, False, Endpoint_Method.POST, "games/game"
    """
    Creates a new game by giving its bggId (BoardGameGeek game id) or a game name. [POST /games/game](https://rankade.com/api/#post-games-game)
    """
    GAMES = True, False, Endpoint_Method.GET, "games"
    """
    Returns the list of the group's games (i.e. games with at least one match played within the group). [GET /games](https://rankade.com/api/#get-games)
    """
    GAMES_POPULAR = True, False, Endpoint_Method.GET, "games/popular"
    """
    Returns the list of rankade's most popular games. [GET /games/popular](https://rankade.com/api/#get-games-popular)
    """
    GAMES_SEARCH = True, False, Endpoint_Method.GET, "games/search"
    """
    Searches games by name. [GET /games/search](https://rankade.com/api/#get-games-search)
    """
    MATCH = True, False, Endpoint_Method.POST, "matches/match"
    """
    Adds one or more matches to the group. [POST /matches/match](https://rankade.com/api/#post-matches-match)
    """
    MATCH_EXISTS = True, False, Endpoint_Method.GET, "matches/match/exists"
    """
    Check if an accepted match exists by giving its external identifier. [GET /matches/match/exists](https://rankade.com/api/#get-matches-match-exists)
    """
    MATCH_STATUS = True, False, Endpoint_Method.GET, "matches/status"
    """
    Retrieves API-recorded matches status. [GET /matches/status](https://rankade.com/api/#get-matches-status)
    """
    MATCHES = True, True, Endpoint_Method.GET, "matches{subset}{page}"
    """
    Retrieve registered and processed matches. [GET /matches](https://rankade.com/api/#get-matches)
    """
    PLAYER = True, False, Endpoint_Method.POST, "players/player"
    """
    Create a ghost player within the group.[POST /players/player](https://rankade.com/api/#post-players-player)
    """
    PLAYERS = True, True, Endpoint_Method.GET, "players{page}"
    """
    Returns the list of the group's players. [GET /player](https://rankade.com/api/#get-players)
    """
    QUOTA = True, False, Endpoint_Method.GET, "quota"
    """
    Returns quota usage details in percentage for the current client. [GET /quota] (https://rankade.com/api/#get-quota)
    """
    RANKINGS = True, True, Endpoint_Method.GET, "rankings{subset}{match}{page}"
    """
    Get a groups rankings, filter by subset, beginning with match number. [GET /rankings](https://rankade.com/api/#get-rankings)
    """
    STATUS = True, False, Endpoint_Method.GET, "status"
    """
    Returns the status of the service. [GET /status](https://rankade.com/api/#get-status)
    """
    SUBSET = True, False, Endpoint_Method.GET, "subset"
    """
    Returns a list of the groups subsets. [GET /subsets](https://rankade.com/api/#get-subsets)
    """
    _TEST = False, False, Endpoint_Method.GET, "test"
    """
    Endpoint used for unit tests.
    """


@dataclass
class Endpoint_Request:
    """Full information needed for Api.request to make a request."""

    endpoint: Endpoint
    """Basic information for Endpoint"""
    params: PARAMS = field(default=None)
    """Additional Parameters to be used in request"""
    headers: HEADERS = field(default=None)
    """Additional Headers to be used in the request"""
    json: JSON = field(default=None)
    """JSON to be sent as part of the request"""
    page: Optional[int] = field(init=False, default=None)
    """Page number to be requested"""
    match: Optional[int] = field(default=None)
    """Match number to be requested"""
    subset: Optional[str] = field(default=None)
    """Subset Id to be requested"""

    def __post_init__(self):
        if self.endpoint.paginated:
            self.page = 1

    @property
    def path(self) -> str:
        """Builds the URL path to use, from the base path included in the Endpoint with {page}, {match}, & {subset} replaced with the appropriate value"""
        page_str = ""
        subset_str = ""
        if self.page is not None:
            page_str = "/" + str(self.page)
        if self.match is None:
            match_str = "/last"
        else:
            match_str = "/" + str(self.match)
        if self.subset is not None:
            subset_str = "/" + self.subset
        return self.endpoint.base_path.format(subset=subset_str, match=match_str, page=page_str)

    @property
    def method(self) -> str:
        """Returns the required HTTP method to be used for the endpoint."""
        return self.endpoint.method.value

    @property
    def requires_auth(self) -> bool:
        """Does the endpoint require authorisation before making the request."""
        return self.endpoint.auth

    @property
    def is_paginated(self) -> bool:
        """Will the request return a paginated response."""
        return self.endpoint.paginated

    def add_parameter(self, name: str, value: str) -> MutableMapping[str, Union[str, int]]:
        """Convenience method to add a custom parameter."""
        if self.params is None:
            self.params = {}
        self.params[name] = value
        return self.params

    def add_header(self, name: str, value: str) -> MutableMapping[str, str]:
        """Convenience method to add a custom header."""
        if self.headers is None:
            self.headers = {}
        self.headers[name] = value
        return self.headers

    def set_json(self, json: JSON) -> JSON:
        """Convenience method to set the JSON body of the request."""
        if self.json is None:
            self.json = {}
        self.json = json
        return self.json



# rankade.Rankade.py
from typing import MutableSequence, Optional, Union

import aiohttp

import rankade.models as models
from rankade.api import Api, Endpoint
from rankade.api.Api import PARAMS


class Rankade(object):
    """Main wrapper around the Rankade Api. Use this class along with it's methods to access the Api."""

    def __init__(
        self,
        key_or_token: str,
        secret: Optional[str] = None,
        base_url: Optional[str] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        self._api = Api(key_or_token, secret, base_url=base_url, session=session)

    """
    Games
    """

    async def get_games(self) -> models.Games:
        """Returns the list of the group's games (i.e. games with at least one match played within the group)."""
        async with self._api as api:
            games_response = await api.request(endpoint=Endpoint.GAMES)
            return models.Games.from_dict(data_dict=games_response)

    async def get_popular_games(self) -> models.Games:
        """
        Not implemented on server, is in API documentation, but requests return an error.

        :raises: NotImplementedError
        """
        raise NotImplementedError("Not implimented on server, request returns an error.")

        # If ever gets implimented on server this should be uncommented.
        # async with self._api as api:
        #   games_attributes = await self._api.request(endpoint=Endpoint.POPULAR)
        #   return models.Games.from_dict(data_dict=games_attributes)

    async def game_search(self, name: str) -> models.Games:
        """
        Searches games by name.

        :param name: Name of game to be searched for
        """
        async with self._api as api:
            search_endpoint = Endpoint.GAMES_SEARCH
            params: PARAMS = {"name": name}
            games_response = await api.request(endpoint=search_endpoint, params=params)
            return models.Games.from_dict(data_dict=games_response)

    async def new_game_with_bggId(self, bggId: int) -> models.Game:
        params: PARAMS = {"bggId": bggId}
        return await self._new_game_with(params=params)

    async def new_game_with_name(self, name: str) -> models.Game:
        params: PARAMS = {"name": name}
        return await self._new_game_with(params=params)

    async def _new_game_with(self, params: PARAMS) -> models.Game:
        async with self._api as api:
            result = await api.request(endpoint=Endpoint.GAME, params=params)
            games = models.Games.from_dict(data_dict=result)
            return games[0]

    """
    Matches
    -------
    """

    async def save_match(self, match: models.NewMatch, dry_run: bool = False) -> models.NewMatchResponse:
        """
        To save a match pass in the following parameters:
        :params models.NewMatch match: Completed new match model.
        :params bool dry_run: For testing, allows for debugging without actually saving the match or using any of the groups Quota.
        """
        params: dict[str, Union[str, int]] = {}
        if dry_run:
            params = {"dryrun": "true"}
        async with self._api as api:
            result = await api.request(endpoint=Endpoint.MATCH, json=match.as_dict(), params=params)
            return models.NewMatchResponse(**result)

    async def get_match_status(self) -> models.MatchStatus:
        """Get the status of matches posted by the API."""
        async with self._api as api:
            status_response = await api.request(endpoint=Endpoint.MATCH_STATUS)
            return models.MatchStatus(**status_response)

    async def get_all_matches(self) -> models.Matches:
        """Get all Matches from the group."""
        async with self._api as api:
            matches_response = await api.request(endpoint=Endpoint.MATCHES)
            return models.Matches.from_dict(data_dict=matches_response)

    async def get_match_with_id(self, id: str) -> Optional[models.Match]:
        """
        Get a match from the group with an Id matching the parameter.
        :param str id: Match Id to find.
        """
        matches = await self.get_all_matches()
        match = [m for m in matches if m.id == id]
        return match[0] if match else None

    async def get_matches_with_players(self, player_ids: MutableSequence[str]) -> Optional[models.Matches]:
        """
        Get Matches player by players matching Id's matching the parameter.
        :param: MutableSequence[str] player_ids: Id's of players to find matches they've player.
        """
        all_matches = await self.get_all_matches()
        filtered_matches = [m for m in all_matches if any(id in m.player_ids for id in player_ids)]
        return models.Matches(data=filtered_matches, totalMatches=len(filtered_matches))

    async def get_match_number(self, number: int) -> Optional[models.Match]:
        """
        Get match by number matching parameter.
        :param int number: Match number to find.
        """
        matches = await self.get_all_matches()
        match = [m for m in matches if m.number == number]
        return match[0] if match else None

    """
    Players
    """

    async def get_all_players(self) -> models.Players:
        """Get a list of all the players in the group."""
        async with self._api as api:
            players_response = await api.request(endpoint=Endpoint.PLAYERS)
            return models.Players.from_dict(data_dict=players_response)

    async def new_ghost_player(self, name: str) -> models.Player:
        """
        Make a new ghost player to add to the group.
        There is a limit on Ghost users see [Quotas and Limits](https://rankade.com/api/#quota-and-limits).
        """
        assert isinstance(name, str)
        async with self._api as api:
            params: PARAMS = {"name": name}
            ghost_response = await api.request(endpoint=Endpoint.PLAYER, params=params)
            return models.Players.from_dict(data_dict=ghost_response)[0]

    """
    Quota
    """

    async def get_quota(self) -> models.Quota:
        """Returns current quota usage percentage"""
        async with self._api as api:
            quota_response = await api.request(endpoint=Endpoint.QUOTA)
            return models.Quota(**quota_response)

    """
    Rankings
    """

    async def get_rankings(
        self, subset_id: Optional[str] = None, match_number: Optional[int] = None
    ) -> models.Rankings:
        """Retrieve group's ranking for selected subset after the selected match number.
        :param subset_id: Id of the subset to be used.
        :param match_number: Retrieve subset results starting at match number provided.
        """
        assert isinstance(subset_id, (str, type(None)))
        assert isinstance(match_number, (int, type(None)))
        async with self._api as api:
            ranking_response = await api.request(endpoint=Endpoint.RANKINGS, subset=subset_id, match=match_number)
            return models.Rankings.from_dict(data_dict=ranking_response)

    """
    Subsets
    """

    async def get_subset_with_id(self, id: str) -> Optional[models.Subset]:
        """Return a specific subset using it's Id."""
        subsets = await self.get_subsets()
        subset = [s for s in subsets if s.id == id]
        return subset[0] if subset else None

    async def get_subsets(self) -> models.Subsets:
        """Return a list of all of a groups subsets."""
        async with self._api as api:
            subsets_attributes = await api.request(endpoint=Endpoint.SUBSET)
            return models.Subsets.from_dict(data_dict=subsets_attributes)



# rankade.models.Player.py
from dataclasses import dataclass, field
from typing import ClassVar, Dict, List, Optional, Type

from ..Consts import GHOST_PREFIX
from .Base import Page, RankadeObject


@dataclass(kw_only=True, slots=True)
class Player(RankadeObject):
    """Represents a single player returned by the Rankade API."""

    id: str
    """Player id"""
    ghost: bool
    """True if player is a ghost"""
    displayName: str
    """Player display name"""
    icon: str
    """Player Icon URL"""
    username: Optional[str] = None
    """Player username, will be None if player is a ghost"""

    def __post_init__(self):
        if isinstance(self.ghost, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.ghost = bool(self.ghost)

    @property
    def is_ghost(self):
        """Convenience method returns self.ghost"""
        return self.ghost


@dataclass(kw_only=True, slots=True)
class Players(Page[Player]):
    """
    Represents a list of player objects returned by the Rankade server.
    Individual player objects returned by the server can be accessed in the same way as a regular list.
    """

    totalPlayers: int = field(default=0)
    """Total matches on all pages."""
    _content_class: ClassVar[Type[RankadeObject]] = Player
    """Classvar to allow the an object in the list to be created from a dict returned from the server."""

    @property
    def ids(self) -> List[str]:
        """Player & ghost ids"""
        return [player.id for player in self.data]

    @property
    def ghosts(self) -> List[Player]:
        """All ghost players."""
        return [player for player in self.data if player.is_ghost]

    @property
    def display_names(self) -> List[str]:
        """
        Display names for players & ghosts.
        Ghost display names have a '*' prefix.
        """
        return [player.displayName for player in self.data]

    @property
    def display_names_clean(self) -> List[str]:
        """
        Display names for players & ghosts.
        Ghost display names will be returned without '*' prefix.
        """
        return [player.displayName.replace(GHOST_PREFIX, "") for player in self.data]

    @property
    def usernames(self) -> List[str | None]:
        """Usernames for all players, ghosts not included."""
        x = [player.username for player in self.data if player.username != ""]
        return x

    @property
    def icons(self) -> Dict[str, str]:
        """All icon URIs for players & ghosts (if set.)"""
        ids = [player.id for player in self.data if player.icon != ""]
        icons = [player.icon for player in self.data if player.icon != ""]
        return dict(zip(ids, icons, strict=False))



# rankade.models.Subset.py
from dataclasses import dataclass
from datetime import datetime
from typing import ClassVar, Optional, Type

from .Base import RankadeObject, ResultList
from .Game import Game
from .Match import Match, Matches


@dataclass(kw_only=True, slots=True)
class Subset(RankadeObject):
    """Represents a groups subset."""

    id: str
    """Subset id."""
    name: str
    """Subset name."""
    type: str
    """Subset type."""
    creationDate: datetime
    """Subset creation date."""
    isMain: bool
    """Subset is the main subset."""
    isCustom: bool
    """Subset is a custom subset."""
    icon: str
    """Subset icon URL."""
    countMatches: int
    """Total matches in subset."""
    game: Optional[Game]
    """Game associated with the subset."""
    firstMatch: Match
    """First match included in the subset."""
    lastMatch: Match
    """Last match included in the subset."""

    def __post_init__(self):
        if isinstance(self.isMain, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.isMain = bool(self.isMain)
        if isinstance(self.isCustom, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.isCustom = bool(self.isCustom)
        if isinstance(self.creationDate, str):
            self.creationDate = datetime.fromisoformat(self.creationDate)
        if isinstance(self.game, dict):
            self.game = Game(**self.game)  # pyright: ignore[reportUnknownArgumentType]

        if isinstance(self.firstMatch, dict):
            self.firstMatch = Match(**self.firstMatch)  # pyright: ignore[reportUnknownArgumentType]

        if isinstance(self.lastMatch, dict):
            self.lastMatch = Match(**self.lastMatch)  # pyright: ignore[reportUnknownArgumentType]

    @property
    def matches(self):
        """Includes bot the first and last match in the subset."""
        return Matches(data=[self.firstMatch, self.lastMatch], totalMatches=2)


@dataclass(kw_only=True, slots=True)
class Subsets(ResultList[Subset]):
    """
    Represents the list of subsets returned by the Rankade server.
    Individual subset objects returned by the server can be accessed in the same way as a regular list.
    """

    _content_class: ClassVar[Type[RankadeObject]] = Subset
    """Classvar to allow the an object in the list to be created from a dict returned from the server."""



# rankade.models.__init__.py
from .Base import RankadeObject, ResultList
from .Error import Error, Errors
from .Faction import Faction, Factions
from .Game import Game, Games
from .Match import Match, Matches
from .NewMatch import NewMatch, NewMatchList
from .NewMatchResponse import (
    MatchStatus,
    NewMatchResponse,
    NewMatchReturn,
    NewMatchReturnList,
)
from .Player import Player, Players
from .Quota import Quota
from .Ranking import Ranking, Rankings
from .Subset import Subset, Subsets

__all__ = [
    "ResultList",
    "Error",
    "Errors",
    "Faction",
    "Factions",
    "Game",
    "Games",
    "Match",
    "Matches",
    "MatchStatus",
    "NewMatch",
    "NewMatchList",
    "NewMatchResponse",
    "NewMatchReturn",
    "NewMatchReturnList",
    "Player",
    "Players",
    "Quota",
    "RankadeObject",
    "Ranking",
    "Rankings",
    "ResultList",
    "Subset",
    "Subsets",
]



# rankade.models.Ranking.py
from dataclasses import dataclass
from typing import ClassVar, Dict, Type

from .Base import Page, RankadeObject
from .Match import Match
from .Player import Player
from .Subset import Subset


@dataclass(kw_only=True, slots=True)
class Ranking(RankadeObject):
    """Individual Ranking of a player in the subset."""

    ree: int
    """Player's Ree score. (In the subset or after selected match.)"""
    deltaRee: int
    """Player's change in Ree score. (In the subset or after selected match.)"""
    position: int
    """Player's position. (In the subset or after selected match.)"""
    deltaPosition: int
    """Players change in position. (In the subset or after selected match.)"""
    belt: int
    """Players current belt."""
    beltLabel: str
    """Label of players current belt."""
    title: int
    """Players Rank title."""
    titleLabel: str
    """Players Rank title."""
    status: int
    """Players status."""
    statusLabel: str
    """Players status."""
    player: Player
    """Player object"""

    def __post_init__(self):
        if isinstance(self.player, Dict):
            self.player = Player(**self.player)  # pyright: ignore[reportUnknownArgumentType]


@dataclass(kw_only=True, slots=True)
class Rankings(Page[Ranking]):
    """
    Represents the list of Rankings returned by the Rankade server.
    Individual ranking objects returned by the server can be accessed in the same way as a regular list.
    """

    _content_class: ClassVar[Type[RankadeObject]] = Ranking
    """Classvar to allow the an object in the list to be created from a dict returned from the server."""

    match: Match
    """Match after which the rankings were calculated"""
    subset: Subset
    """Subset which the ranking applies to."""

    def __post_init__(self):
        if isinstance(self.match, Dict):
            self.match = Match(**self.match)  # pyright: ignore[reportUnknownArgumentType]

        if isinstance(self.subset, Dict):
            self.subset = Subset(**self.subset)  # pyright: ignore[reportUnknownArgumentType]

        self.data.sort(key=lambda position: position.position)

    @property
    def sorted_by_position(self):
        """Rankings sorted by position."""
        r = self.data[:]
        r.sort(key=lambda position: position.position)
        return r

    @property
    def sorted_by_delta_position(self):
        """Rankings sorted by change of position."""
        r = self.data[:]
        r.sort(reverse=True, key=lambda position: position.deltaPosition)
        return r

    @property
    def sorted_by_ree(self):
        """Rankings sorted by Ree score."""
        r = self.data[:]
        r.sort(reverse=True, key=lambda position: position.ree)
        return r

    @property
    def sorted_by_delta_ree(self):
        """Rankings sorted by change of Ree score."""
        r = self.data[:]
        r.sort(reverse=True, key=lambda position: position.deltaRee)
        return r



# rankade.api.RankadeResponse.py
from dataclasses import dataclass
from typing import Any, MutableMapping, Optional


@dataclass(slots=True)
class RankadeResponse(object):
    """First response from server, should never leave the Api class and be surfaced to user of the package."""

    success: Optional[MutableMapping[str, Any]] = None
    errors: Optional[MutableMapping[str, Any]] = None



# rankade.models.Game.py
from dataclasses import dataclass, field
from typing import ClassVar, Optional, Type

from .Base import RankadeObject, ResultList


@dataclass(kw_only=True, slots=True)
class Game(RankadeObject):
    """Represents a single Game object returned by Rankade"""

    id: int
    """Rankade id of the game."""
    name: str
    """ Game name."""
    weight: str
    """ Should be one of ultralight, light, midlight, normal, heavy, massive.
    A heavier game will result in a larger variation of the ree score."""
    weightLabel: str
    """Appears to be a nicely formatted version of the weight attribute."""
    mediumImage: str
    """URL for game art."""
    thumbnail: str
    """URL for game art thumbnail."""
    bggIdGame: Optional[int] = field(default=None)
    """Board Game Geek id for the game."""


@dataclass(kw_only=True, slots=True)
class Games(ResultList[Game]):
    """
    Represents the list of Games returned by the Rankade server.
    Individual game objects returned by the server can be accessed in the same way as a regular list.
    """

    _content_class: ClassVar[Type[RankadeObject]] = Game
    """Classvar to allow the an object in the list to be created from a dict returned from the server."""



# rankade.__init__.py
__title__ = "rankade"
__author__ = "14zombies"
__license__ = "MIT"
__copyright__ = "Copyright 2021-present 14zombies"
__version__ = "0.1.6"

import logging

from rankade import RankadeExceptions
from rankade.models import *
from rankade.Rankade import Rankade

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "Rankade",
    "RankadeExceptions",
    "Error",
    "Errors",
    "Faction",
    "Factions",
    "Game",
    "Games",
    "Match",
    "Matches",
    "MatchStatus",
    "NewMatch",
    "NewMatchList",
    "NewMatchResponse",
    "NewMatchReturn",
    "NewMatchReturnList",
    "Player",
    "Players",
    "Quota",
    "RankadeObject",
    "Ranking",
    "Rankings",
    "ResultList",
    "Subset",
    "Subsets",
]



# rankade.RankadeExceptions.py
class RankadeException(Exception):
    """
    Base Exception, should not be raised directly.
    """

    def __init__(self, message: str):
        self.message = message


class ApiErrorResponse(RankadeException):
    """
    Base class for error response from API.
    Raised when no appropriate ApiErrorResponse subclass.

    Raised from:
    - {py:meth}`rankade.api.Api.Api._check_request`

    :param str url: URL queried.
    :param str verb: HTTP Method used.
    :param str status: Request HTTP status code.
    :param str code: API error code.
    :param str message: API error message.
    """

    def __init__(self, url: str, verb: str, status: int, code: str = "", message: str = ""):
        self.url = url
        self.verb = verb
        self.status = status
        self.code = code
        super().__init__(message)


class AuthCredentials(ApiErrorResponse):
    """
    Code will be "A001" or "A002", raised when key and/or secret are
    invalid or access has been disabled.

    Raised from:
    - {py:meth}`rankade.api.Api.Api._check_request`
    """

    pass


class MatchValidation(ApiErrorResponse):
    """
    Code will begin with "M" and message will contain
    contain details of validation error with the match.

    Raised from:
    - {py:meth}`rankade.api.Api.Api._check_request`
    """

    pass


class NoValidCredentials(RankadeException):
    """
    Raised when either no key & secret or no token has been supplied
    to main Rankade class.

    Raised from:
    - {py:class}`rankade.api.Api.Api`
    - {py:meth}`rankade.api.Api.Api._request_jwt`
    """

    def __init__(self, message: str = "No Credentials Supplied"):
        super().__init__(message)


class Quotas(ApiErrorResponse):
    """
    Code will begin with "Q" and message will contain
    details of the quota limit reached. Status should be either 429 or 202.

    Raised from:
    - {py:meth}`rankade.api.Api.Api._check_request`
    """

    pass


class SearchTooShort(RankadeException):
    """
    Raised when search term is less than 2 characters.
    API will return missing or invalid parameters which is not useful.

    Raised from:
    - {py:meth}`rankade.Rankade.Rankade.game_search`
    """

    def __init__(self, search: str = ""):
        self.search = search
        super().__init__(f"Search must be at least 2 characters (got '{search}').")



# rankade.models.Error.py
from dataclasses import dataclass, field
from typing import ClassVar, Type

from .Base import RankadeObject, ResultList


@dataclass(kw_only=True, slots=True)
class Error(RankadeObject):
    """
    Represents a single error returned by the Rankade API.

    :::{seealso}
    List of codes & messages returned can be found in the Rankade API documentation.
        - [Error Responses](https://rankade.com/api/#error-responses)
        - [Quota and Limits](https://rankade.com/api/#quota-and-limits)
        - [Get Auth](https://rankade.com/api/#get-auth)
        - [Post Players Player](https://rankade.com/api/#post-players-player)
        - [Post Matched Match](https://rankade.com/api/#post-matches-match)
    :::

    :param str message: Error message returned by the server.
    :param str code: Error code of returned error.

    """

    code: str
    """Error code in returned error.
        If first character will tell you what type of error it is:
            - A: Auth
            - Q: Quota
            - M: Match Validation
    """
    message: str
    """Error message returned by the server."""


@dataclass(kw_only=True, slots=True)
class Errors(ResultList[Error]):

    """
    Represents a list of error objects returned by the Rankade server.
    Individual error objects returned by the server can be accessed in the same way as a regular list.

    :param str url: Queried URL that provided the error response.
    :param str verb: HTTP method used in query (GET, POST, etc)
    :param int status: HTTP status code returned with the error response.
    :param MutableMapping[str, Any] data: List of errors returned in json dict.


    :::{note}
    I have only ever received a single error.
    As the spec is ambiguous this class caters for more than one error.
    :::

    """

    _content_class: ClassVar[Type[RankadeObject]] = Error

    url: str = field(default="")
    """Queried url that provided the error response."""
    verb: str = field(default="")
    """HTTP method used in query (GET, POST, etc)"""
    status: int = field(default=0)
    """HTTP status code returned with the error response."""



# rankade.models.Match.py
from dataclasses import dataclass
from datetime import datetime
from typing import ClassVar, Dict, List, Type

from .Base import Page, RankadeObject
from .Faction import Faction, Factions
from .Game import Game
from .Player import Player, Players


@dataclass(kw_only=True, slots=True)
class Match(RankadeObject):
    weight: str
    weightLabel: str
    notes: str
    id: str
    externalId: str
    date: datetime
    registrationDate: datetime
    number: int
    summary: str
    type: str
    draw: bool
    game: Game
    factions: Factions

    def __post_init__(self):
        if isinstance(self.draw, (str, int)):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.draw = bool(self.draw)
        if isinstance(self.date, str):
            self.date = datetime.fromisoformat(self.date)
        if isinstance(self.registrationDate, str):
            self.registrationDate = datetime.fromisoformat(self.registrationDate)
        if isinstance(self.game, Dict):
            self.game = Game(**self.game)  # pyright: ignore[reportUnknownArgumentType]

        if isinstance(self.factions, (List, Dict)):
            if not isinstance(self.factions[0], Faction):  # pyright: ignore[reportUnnecessaryIsInstance]
                self.factions = Factions.from_dict(data_dict=self.factions)  # pyright: ignore[reportArgumentType]
            self.factions.sort(key=lambda f: f.rank)

    @property
    def is_draw(self) -> bool:
        return bool(self.draw)

    @property
    def winning_factions(self) -> List[Faction]:
        return [faction for faction in self.factions if faction.rank == 1]

    @property
    def winning_players(self) -> List[Player]:
        faction_players = [faction.players for faction in self.winning_factions]
        return [player for faction in faction_players for player in faction]

    @property
    def players(self) -> Players:
        players = [faction.players for faction in self.factions]
        flat_players = [x for xs in players for x in xs]
        return Players(data=flat_players)

    @property
    def player_ids(self) -> List[str]:
        return [p.id for p in self.players]


@dataclass(kw_only=True, slots=True)
class Matches(Page[Match]):

    """
    Represents a list of match objects returned by the Rankade server.
    Individual match objects returned by the server can be accessed in the same way as a regular list.
    """

    totalMatches: int
    """Total matches on all pages."""
    _content_class: ClassVar[Type[RankadeObject]] = Match
    """Classvar to allow the an object in the list to be created from a dict returned from the server."""

    def all_players(self) -> Players:
        """Returns all players from all factions of all matches."""
        players = [player for match_players in self.data for player in match_players.players]
        deduped_players = [player for count, player in enumerate(players) if player not in players[count + 1 :]]
        return Players(data=deduped_players)



# rankade.models.NewMatchResponse.py
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional, Type

from .Base import RankadeObject, ResultList
from .Error import Errors

"""
NewMatchResponse is the object that comes back from the server. It will contain 2 lists "accepted & rejected" which are NewMatchReturnList objects.
The NewMatchReturnList objects are lists of NewMatchReturn.

"""


@dataclass(kw_only=True, slots=True)
class MatchStatus(RankadeObject):
    """Retrieves API-recorded matches status. [Details](https://rankade.com/api/#get-matches-status)

    :::{note}
    Counts are only for matches entered via API.
    Matches entered via webapp/app are not included.
    :::
    """

    added: int
    """Matches added to subset(s), both processed and not yet processed."""
    processed: int
    """Processed matches."""
    queued: int
    """Accepted matches, queued for insertion."""
    total: int
    """Total API-recorded matches."""
    waiting: int
    """Inserted matches, waiting for their subset(s)."""


@dataclass(kw_only=True, slots=True)
class NewMatchReturn(RankadeObject):
    """Individual match status item returned upon submit."""

    index: int
    """Match index."""
    id: str
    """Match Id."""
    name: str
    """Match name."""
    errors: Optional[Errors] = None
    """List of errors in match, if appropriate."""

    def __post_init__(self):
        if isinstance(self.errors, List):
            self.errors = Errors.from_dict(data_dict=self.errors)  # pyright: ignore[reportArgumentType]


@dataclass(kw_only=True, slots=True)
class NewMatchReturnList(ResultList[NewMatchReturn]):
    """lists "accepted & rejected" from NewMAtchResponse are NewMatchReturnList objects."""

    _content_class: ClassVar[Type[RankadeObject]] = NewMatchReturn
    """Classvar to allow the an object in the list to be created from a dict returned from the server."""


@dataclass(kw_only=True, slots=True)
class NewMatchResponse(RankadeObject):
    """Response from the server to a posted Match."""

    total: int
    """Total matches submitted."""
    acceptedCount: int
    """Total matches accepted."""
    rejectedCount: int
    """Total matches rejected."""
    rejected: NewMatchReturnList
    """List of rejected matches with errors."""
    accepted: NewMatchReturnList
    """List of accepted matches."""
    dryrun: bool
    """Reports True if posting was a test and not acutally submitted."""

    def __init__(
        self,
        total: int,
        acceptedCount: int,
        rejectedCount: int,
        rejected: Dict[str, Any],
        accepted: Dict[str, Any],
        dryrun: bool = False,
    ):
        """
        :param int total: Total matches submitted.
        :param int acceptedCount: Total matches accepted.
        :param int rejectedCount: Total matches rejected.
        :param Dict[str, Any] rejected: List of rejected matches with errors.
        :param Dict[str, Any] accepted: List of accepted matches.
        :param bool dryrun: Reports True if posting was a test and not acutally submitted.
        """
        self.total = total
        self.acceptedCount = acceptedCount
        self.rejectedCount = rejectedCount

        self.rejected = NewMatchReturnList.from_dict(data_dict=rejected)
        self.accepted = NewMatchReturnList.from_dict(data_dict=accepted)
        if not isinstance(dryrun, bool):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.dryrun = bool(dryrun)

    @property
    def has_error(self) -> bool:
        """Convenience method returns True if any submitted matches are rejected."""
        return len(self.rejected) > 0



# rankade.models.Base.py
from __future__ import annotations

from abc import ABC
from collections import UserList
from copy import deepcopy
from dataclasses import dataclass, field
from typing import (
    Any,
    ClassVar,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Type,
    TypeVar,
    Union,
)

T = TypeVar("T", bound="RankadeObject")


@dataclass(slots=True)
class RankadeObject(ABC):  # noqa: B024

    """
    Base class for all objects returned from the server.
    All models should inherit from this class.
    """

    pass


@dataclass(kw_only=True, slots=True)
class ResultList(UserList[T], RankadeObject):
    """
    Base class for lists of items from server.

    :param List[rankade.models.Base.T] data: List of RankadeObjects returned by the sever.
    """

    _content_class: ClassVar[Type[RankadeObject]]
    """Classvar to be overridden on each subclass to allow the an object in the list to be created from a dict returned from the server."""
    data: List[T]
    """List of RankadeObjects returned by the sever."""

    @classmethod
    def from_dict(cls, data_dict: Union[MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]]):
        """
        Create a ResultList instance from a dict.
        """
        content_class = cls._content_class
        if isinstance(data_dict, MutableSequence):
            data_dict = {"data": data_dict}
        data = [content_class(**item) for item in data_dict.get("data", [])]
        # make a copy here as it was modifying the original and screwing up tests!
        copyof_data_dict = deepcopy(data_dict)
        del copyof_data_dict["data"]
        return cls(**copyof_data_dict, data=data)


@dataclass(kw_only=True, slots=True)
class Page(ResultList[T]):
    """
    Base class for page of items from server .

    :param List[rankade.models.Base.T] data: List RankadeObject objects.
    :param int page: Current page number.
    :param int totalPages: Total Pages.
    :param int rowsForPage: Max number of items on page.

    """

    page: int = field(default=1)
    """Page number"""
    totalPages: int = field(default=1)
    """Total pages"""
    rowsForPage: Optional[int] = None
    """How many items the server will return on each page."""



# rankade.models.NewMatch.py
from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, Optional, Type, Union

from rankade.api.Api import JSON
from rankade.RankadeExceptions import RankadeException

from ..Consts import BOT_ID
from .Base import RankadeObject, ResultList
from .Faction import Faction, Factions
from .Game import Game
from .Player import Player, Players


@dataclass
class NewMatch(RankadeObject):
    """This class to be used for creating Matches to post to the Rankade API."""

    notes: str
    """General notes."""
    name: str = ""
    """
    Match name.
    :::{note}
    Use any custom string for the name of the match or 2 special options are available:

    "default_1"
        : set game's name as the match name.

    "default_2"
        : to set faction/player name vs other faction/player name as the match name.
    :::
    """
    game: Optional[Game] = field(default=None)
    """Game id retrieved from ` GET /games/search`, `GET /games`, or `GET /games/popular` API calls."""
    factions: Optional[Factions] = field(default=None)
    """Factions' data."""
    weight: str = field(default="normal")
    """Match weight, chosen from this list: ultralight, light, midlight, normal, heavy, massive.
    A heavier game results in larger variations in ree score, a lighter game in smaller ones. We strongly suggest you to set weights in a way that most of the matches have normal weight. See FAQ for details."""

    def add_bot_faction(self, rank: int, points: str, name: str = ""):
        """
        Convenience method to add a bot faction to the match.
        """
        bot_player = Players(data=[Player(id=BOT_ID, ghost=False, displayName=BOT_ID, icon="")])
        self.add_faction(players=bot_player, rank=rank, points=points, name=name, bot=True)

    def add_faction(self, players: Union[Players, Player], rank: int, points: str, name: str = "", bot: bool = False):
        """
        Convenience method to add a faction to the match.
        """
        if isinstance(players, Player):
            players = Players(data=[players])
        if len(players) < 1:
            raise RankadeException(message="A faction must contain at least 1 player")
        bot_players = [player for player in players if player.id == BOT_ID]
        if len(bot_players) > 1 or (bot and len(players) > 1):
            raise RankadeException(message="A bot faction cannot contain any other players")

        faction = Faction(points=points, rank=rank, players=players, name=name, bot=bot)
        if self.factions is None:
            self.factions = Factions(data=[faction])
        else:
            self.factions.append(faction)

    def as_dict(self) -> JSON:
        """
        Returns a custom dictionary for new match server post request.
        """
        if self.game is None or self.factions is None:
            return []
        factions = [f.as_dict() for f in self.factions]
        match: JSON = {
            "name": self.name,
            "game": self.game.id,
            "weight": self.weight or self.game.weight,
            "factions": factions,
            "notes": self.notes,
        }
        return [match]


@dataclass
class NewMatchList(ResultList[NewMatch]):
    """
    Represents the list of NewMatches returned by the Rankade server.
    Individual NewMatch objects returned by the server can be accessed in the same way as a regular list.
    """

    _content_class: ClassVar[Type[RankadeObject]] = NewMatch
    """ClassVar to allow the an object in the list to be created from a dict returned from the server."""



# rankade.api.Token.py
import logging
from dataclasses import dataclass, field
from typing import List, Optional

import jwt

logger = logging.getLogger(__name__)


@dataclass(kw_only=True, slots=True)
class Token:
    """
    JWT Object

    """

    token: str
    """JWT string"""
    algorithms: Optional[List[str]] = field(default=None)
    """List of JWA Algorithms to use in JWT decoding.
    See [rfc7518 section-3.1](https://datatracker.ietf.org/doc/html/rfc7518#section-3.1)
    for a full list. "HS256" by default."""

    def __post_init__(self):
        if self.algorithms is None:
            self.algorithms = ["HS256"]

    @property
    def bearer(self) -> str:
        """For use with authorisation returns "Bearer" + token."""
        return f"Bearer {self.token}"

    @property
    def is_invalid(self) -> bool:
        """Checks whether the token is invalid or not. By checking if the token can be decoded then checks the expiry date."""
        try:
            jwt.decode(self.token, options={"verify_signature": False, "verify_exp": True}, algorithms=self.algorithms)  # pyright: ignore[reportUnknownMemberType] #
            logger.info("🎟 Token is valid")
            return False
        except jwt.ExpiredSignatureError:
            logger.info("☠️ Token is expired")
            return True
        except jwt.PyJWTError as exp:
            logger.critical(f"⛔️ Token error - {exp}")
            return True



# rankade.models.Faction.py
from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar, Type

from rankade.api.Api import JSON

from ..Consts import BOT_ID
from .Base import RankadeObject, ResultList
from .Player import Players


@dataclass(kw_only=True, slots=True)
class Faction(RankadeObject):
    """Represents a faction used within a Match."""

    name: str
    """Faction Name"""
    points: str
    """Faction points/score,"""
    rank: int
    """Faction ranking compared to other factions."""
    bot: bool = False
    """returns true if faction is bot faction."""
    winner: bool = False
    """returns true if faction is winning faction."""
    players: Players
    """List of faction players."""
    countPlayers: int = 0
    """Number of players in the faction."""

    def __post_init__(self):
        if isinstance(self.bot, int) or self.players[0].id == BOT_ID:  # pyright: ignore[reportUnnecessaryIsInstance]
            self.bot = bool(self.bot)
        if isinstance(self.winner, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.winner = bool(self.winner)
        if not isinstance(self.players, Players):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.players = Players.from_dict(data_dict=self.players)
        self.countPlayers = len(self.players)

    @property
    def is_bot(self) -> bool:
        """Convenience attribute - returns true if faction is bot faction."""
        return self.bot

    @property
    def is_winner(self) -> bool:
        """Convenience attribute - returns true if faction is winning faction."""
        return self.winner

    def as_dict(self) -> JSON:
        """
        Returns a custom dictionary for new match server post request.
        """
        ids: JSON = [str(p.id) for p in self.players]
        faction_dict: JSON = {
            "rank": int(self.rank),
            "score": str(self.points),
            "players": ids,
        }
        return faction_dict


@dataclass(kw_only=True, slots=True)
class Factions(ResultList[Faction]):
    """
    Represents a list of faction objects returned by the Rankade server.
    Individual faction objects returned by the server can be accessed in the same way as a regular list.
    """

    _content_class: ClassVar[Type[RankadeObject]] = Faction
    """Classvar to allow the an object in the list to be created from a dict returned from the server."""



# rankade.api.__init__.py
from .Api import HEADERS, JSON, PARAMS, Api
from .Endpoint import Endpoint, Endpoint_Method, Endpoint_Request
from .RankadeResponse import RankadeResponse
from .Token import Token

__all__ = [
    "Api",
    "Endpoint",
    "Endpoint_Method",
    "Endpoint_Request",
    "HEADERS",
    "JSON",
    "PARAMS",
    "RankadeResponse",
    "Token",
]



# rankade.models.Quota.py
from dataclasses import dataclass

from .Base import RankadeObject


@dataclass(kw_only=True, slots=True)
class Quota(RankadeObject):
    """
    Rankade Quota usage. All items returned as a percentage of total used.
    :::{seealso}
    Quota limits depend on the subscription tier full list can be found on the [Quota and limits](https://rankade.com/api/#quota-and-limits) section of the Rankade API page.
    """

    callsPerYear: str
    """API calls per year"""
    callsPerHour: str
    """API calls per hour"""
    matchesPerYear: str
    """Matches per year"""
    matchesPerDay: str
    """Matches per day"""
    matchesPerHour: str
    """Matches per hour"""
    rankingCallsPerYear: str
    """Ranking calls per year"""
    rankingCallsPerDay: str
    """Ranking calls per day"""
    rankingCallsPerHour: str
    """Ranking calls per hour"""
    apiCreatedGames: str
    """New API-created games"""



# rankade.consts.py
"""Any magic strings to go here."""
DEFAULT_BASE_URL = "https://api.rankade.com/public/api/1/"
"""Rankade server URL."""
BOT_ID = "bot"
"""Bot users have an ID of "bot"."""
GHOST_PREFIX = "*"
"""Ghost players have a display name prefixed with "*"."""
ERROR_STATUS = [400, 401, 403, 404, 429, 500]
"""All Error HTTP Status"""
RANKADE_ERROR_PREFIX = "R"
"""If Rankade server returns 400, 404, or 500, it should also return an error code beginning with "R"."""
MATCH_ERROR_STATUS = [202, 400]
"""Match validation errors will return either 202 or 400 HTTP status code."""
MATHC_ERROR_PREFIX = "M"
"""Match validation errors will return an error code beginning with "M"."""
AUTH_ERROR_STATUS = [401, 403]
"""Authentication errors will return either 202 or 400 HTTP status."""
AUTH_ERROR_PREFIX = "A"
"""Authentication errors will return an error code beginning with "A"."""
QUOTA_ERROR_STATUS = [202, 429]
"""Quota validation errors will return either 202 or 429 HTTP status."""
QUOTA_ERROR_PREFIX = "Q"
"""Quota errors will return an error code beginning with "Q"."""



# rankade.api.Api.py
from __future__ import annotations

import json
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, MutableMapping, Optional, TypeAlias, Union

import aiohttp

from .. import RankadeExceptions
from ..Consts import (
    AUTH_ERROR_PREFIX,
    AUTH_ERROR_STATUS,
    DEFAULT_BASE_URL,
    MATCH_ERROR_STATUS,
    MATHC_ERROR_PREFIX,
    QUOTA_ERROR_PREFIX,
    QUOTA_ERROR_STATUS,
)
from ..models import Errors
from .Endpoint import Endpoint, Endpoint_Request
from .RankadeResponse import RankadeResponse
from .Token import Token

logger = logging.getLogger(__name__)

JSON: TypeAlias = Dict[str, "JSON"] | List["JSON"] | str | int | float | bool | None
"""TypeAlias = Dict[str, "JSON"] | List["JSON"] | str | int | float | bool | None"""

HEADERS: TypeAlias = Dict[str, str] | None
"""TypeAlias = Dict[str, str] | None"""

PARAMS: TypeAlias = Dict[str, str | int] | None
"""TypeAlias = Dict[str, str | int] | None"""


@dataclass()
class Api(object):
    """Wrapper around aiohttp.

    It manages the aiohttp ClientSession, Authorisation, and requests to the Rankade API.
    It is generally best to use this as a context manager as it will manage the creation and cleanup of the ClientSession.

    ```python
        async with Api(...) as api:
            result = api.request(...)
    ```
    It can also be used without the context manager, if you are managing the ClientSession & error handling yourself.

    ```python
        api = Api(...)
        async api.start()
        async api.request(...)
        ...
        async api.stop()
    ```
    """

    _base_url: str
    """Scheme and Resource to send requests to. Defaults to "https://api.rankade.com/public/api/1/"""
    _key: Optional[str] = None
    """Rankade key."""
    _secret: Optional[str] = None
    """Rankade secret."""
    _session: Optional[aiohttp.ClientSession] = None
    """aiohttp ClientSession to use, will create one if not passed in.."""
    _token: Optional[Token] = field(init=False, default=None)
    """Rankade issued token"""

    def __init__(
        self,
        key_or_token: str,
        secret: Optional[str] = None,
        base_url: Optional[str] = None,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        """
        :param str key_or_token: Rankade key or token.
        :param Optional[str] secret: Rankade secret.
        :param Optional[str] base_url: Scheme and Resource to send requests to. Defaults to "https://api.rankade.com/public/api/1/"
        :param Optional[aiohttp.ClientSession] session: aiohttp ClientSession to use, will create one if not passed in.
        """
        if not key_or_token or (not isinstance(secret, (str, type(None)))):
            raise RankadeExceptions.NoValidCredentials()
        assert isinstance(base_url, (str, type(None)))
        if secret is not None:
            self._key = key_or_token
            self._secret = secret
        else:
            self.token = key_or_token
        self._base_url = base_url or DEFAULT_BASE_URL
        self._session = session

    @property
    def _credentials_params(self):
        """A dictionary containing the key and secret for Rankade API for use as parameters in the http request."""
        return {"key": self._key, "secret": self._secret}

    @property
    def token(self) -> Optional[Token]:
        """Representation of a JWT."""
        return self._token

    @token.setter
    def token(self, value: Union[str, Token, None]) -> None:
        if isinstance(value, str):
            self._token = Token(token=value)
        else:
            self._token = value

    def _make_url_for(self, endpoint_path: str = "") -> str:
        """
        Returns combined base_url + path for endpoint
        :param str endpoint_path: endpoint path to be turned into full URL.
        """
        endpoint_path = endpoint_path.format(page="", subset="")
        if self._base_url[-1] != "/":
            self._base_url = self._base_url + "/"
        return self._base_url + endpoint_path

    async def _request_jwt(self) -> Optional[Token]:
        """
        Checks whether the token is valid or set and the will grab a new one using the key & secret provided during initialisation.
        :raises NoValidCredentials: If only a token is passed as a parameter to the constructor.
        """
        if self.token is None or self.token.is_invalid:
            if not isinstance(self._key, str) or not isinstance(self._secret, str):
                raise RankadeExceptions.NoValidCredentials("Unable to renew token as no valid key or secret provided.")
            logger.debug("🎟 Requesting new token")
            auth_endpoint = Endpoint_Request(Endpoint.AUTH)
            auth_endpoint.add_parameter("key", self._key)
            auth_endpoint.add_parameter("secret", self._secret)
            token_response = await self._request(endpoint=auth_endpoint)
            self.token = Token(**token_response)

        return self.token

    async def _paginated_request(self, endpoint: Endpoint_Request) -> MutableMapping[str, Any]:
        """
        Loops through the pages of response from the endpoint and combines them all into one Page to return.
        :param Endpoint_Request endpoint: Full information for the request contained in endpoint.Endpoint

        """

        first_page = await self._request(endpoint)
        last_page_no = first_page.get("totalPages", 1)
        # range 2 as already got the first page
        for i in range(2, last_page_no + 1):
            endpoint.page = i
            first_page["page"] = i

            next_page = await self._request(endpoint)
            first_page.get("data", []).extend(next_page.get("data", []))

        return first_page

    async def _request(self, endpoint: Endpoint_Request) -> MutableMapping[str, Any]:
        """
        Main wrapper around aiohttp.ClientSession.request, takes in the Endpoint request transforms it into the appropriate parameters for the ClientSession request.
        :params Endpoint_Request endpoint: full endpoint request.
        """
        if self._session is None:
            raise RankadeExceptions.RankadeException(message="No session set in Api.")
        response = None
        url = self._make_url_for(endpoint.path)
        logger.debug(f"Fetching from: {url}")
        if endpoint.requires_auth:
            self.token = await self._request_jwt()
        if self.token is not None:
            endpoint.add_header("Authorization", self.token.bearer)

        async with self._session.request(
            method=endpoint.method,
            url=url,
            params=endpoint.params,
            headers=endpoint.headers,
            json=endpoint.json,
        ) as raw_response:
            try:
                logger.debug(f"Made request {endpoint} {endpoint.page or ""}")
                response = RankadeResponse(**await raw_response.json())

                # This is response["success"] XOR response["errors"]
                # 1 should be None, if both or neither are then the response is bad.
                assert response.success is not None
                assert response.success != response.errors

                return response.success
            except AssertionError as exp:
                raise await self._exception_of_last_resort(
                    raw_response, message="Unexpected response from Server."
                ) from exp
            except aiohttp.ContentTypeError as exp:
                raise await self._exception_of_last_resort(
                    raw_response, message=f"aiohttp Invalid content type: {raw_response.content_type}."
                ) from exp
            except json.JSONDecodeError as exp:
                raise await self._exception_of_last_resort(raw_response, message="JSON Decoding failed.") from exp

    async def _check_request(self, raw_response: aiohttp.ClientResponse):
        """
        Overrides aiohttp.ClientRequest.raise_for_status to allow better error handling.

        :param aiohttp.ClientResponse raw_response: Client response object.
        :raises ApiErrorResponse: If something unexpected with the response has happened.
        :raises MatchValidation: Match rejected by the server, due to some kind of validation error.
        :raises AuthCredentials: Invalid credentials or client disabled.
        :raises Quotas: One of the quote limits has been hit.
        """
        if raw_response.status == 200:
            return
        response = RankadeResponse(**await raw_response.json())
        if response.errors is None:
            raise await self._exception_of_last_resort(
                response=raw_response, message="No errors returned from server, despite error code."
            )

        errors = Errors.from_dict(data_dict=response.errors)
        if not errors:
            raise await self._exception_of_last_resort(response=raw_response, message="Errors response empty.")

        exception = None
        error_code = errors[0].code
        error_message = errors[0].message
        # Match validation errors have either status code 202 or 400 and have a error code beginning with 'M' folowed by 3 digits.
        if raw_response.status in MATCH_ERROR_STATUS and error_code[0] == MATHC_ERROR_PREFIX:
            exception = RankadeExceptions.MatchValidation
        # Auth Errors will return with status 401 or 403 and have an error code beginning with 'A' folowed by 3 digits.
        elif raw_response.status in AUTH_ERROR_STATUS or error_code[0] == AUTH_ERROR_PREFIX:
            exception = RankadeExceptions.AuthCredentials
        # Quota Errors will a return with status 202 or 429 and have an error code beginning with 'A' folowed by 3 digits.
        # 202 is the general code for Quota errors, however 429 is for Quota & Match Validation errors.
        elif raw_response.status in QUOTA_ERROR_STATUS or error_code[0] == QUOTA_ERROR_PREFIX:
            exception = RankadeExceptions.Quotas
        if exception is None:
            raise await self._exception_of_last_resort(response=raw_response)
        raise exception(
            raw_response.method, raw_response.url.human_repr(), raw_response.status, error_code, error_message
        )

    async def _exception_of_last_resort(
        self,
        response: aiohttp.ClientResponse,
        message: str = "An unknown error occured while dealing with the request.",
    ) -> RankadeExceptions.ApiErrorResponse:
        """
        If something has gone badly wrong with parsing the response, this will create and return an exception to be raised.
        :param aiohttp.ClientResponse response: Response which we are having an issue with.
        :param str message: Message to be added to the exception, will also include the raw text of the response.
        """
        return RankadeExceptions.ApiErrorResponse(
            url=response.url.human_repr(),
            verb=response.method,
            status=response.status,
            message=f"{message} {await response.text()}",
        )

    async def request(
        self,
        endpoint: Endpoint,
        headers: HEADERS = None,
        params: PARAMS = None,
        json: JSON = None,
        **kwargs: Any,
    ) -> MutableMapping[str, Any]:
        """
        Public facing interface for API class, will call 1 of 2 methods depending
        if the request is expecting a paginated response or not.

        :param Endpoint endpoint: Contains basic config for the request, including if authorisation is required, will the response be paginated, and what http method & path to use. See {py:class}`Endpoint <rankade.api.Endpoint.Endpoint>` for a full list.
        :param rankade.api.Api.HEADERS headers: A dict representing any headers to be passed, normally this is internal use only for Bearer Auth header.
        :param rankade.api.Api.PARAMS params: Parameters to use with the request.
        :param rankade.api.Api.JSON json: JSON to send to the server.
        :param **kwargs: Used to keep the call site as clean as possible, we can make almost all the requests with some mix
            of the above parameters. There are a few parameters that require passing through to the request. {py:class}`Endpoint_Request <rankade.api.Endpoint.Endpoint_Request>` for a full list.

        """
        endpoint_request = Endpoint_Request(endpoint=endpoint, headers=headers, params=params, json=json, **kwargs)
        if endpoint_request.is_paginated:
            logger.debug("Endpoint is pageinated.")
            return await self._paginated_request(endpoint_request)
        else:
            logger.debug("Endpoint is not pageinated.")
            return await self._request(endpoint_request)

    def start(self):
        """
        Creates a Client session if none was passed through.

        :::{note}
        Should only be called manually if not using API as a context manager.
        :::
        """
        if self._session is None or self._session.closed:
            logger.debug("Creating ClientSession.")
            self._session = aiohttp.ClientSession(raise_for_status=self._check_request)

    async def stop(self):
        """
        Closes the Api session.

        :::{note}
        Should only be called manually if not using API as a context manager.
        :::

        """
        if self._session:
            logger.debug("Closing ClientSession.")
            await self._session.close()

    async def __aenter__(self):
        """Entry handler for context manager."""
        logger.debug("Context started.")
        self.start()
        return self

    async def __aexit__(self, *_):
        """Cleanup handler for context manager."""
        logger.debug("Context finishing.")
        # Do not return 'True' here or exceptions will be supressed.
        # https://docs.python.org/3/reference/datamodel.html?#object.__exit__
        await self.stop()
